""" Generator functions related to trial and test functions and the accumulation loop"""

from dune.perftool.options import get_option

from dune.perftool.generation import (domain,
                                      iname,
                                      pymbolic_expr,
                                      globalarg,
                                      valuearg,
                                      get_global_context_value
                                      )
from dune.perftool.ufl.modified_terminals import ModifiedArgumentDescriptor
from dune.perftool.pdelab import (name_index,
                                  restricted_name,
                                  )
from dune.perftool.pdelab.basis import (evaluate_coefficient,
                                        evaluate_coefficient_gradient,
                                        lfs_iname,
                                        name_basis,
                                        name_lfs_bound,
                                        )
from dune.perftool import Restriction
from pymbolic.primitives import Call, Subscript, Variable

import loopy




def name_trialfunction_gradient(element, restriction, component):
    rawname = "gradu" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)

    # TODO
    #
    # This is just a temporary test used to create an A-matrix as
    # local operator class member. Right now it doesn't evaluate
    # anything.
    if get_option("sumfact"):
        from dune.perftool.sumfact import evaluate_coefficient_gradient_sumfact
        evaluate_coefficient_gradient_sumfact()

    evaluate_coefficient_gradient(element, name, container, restriction, component)
    return name


def name_trialfunction(element, restriction, component):
    rawname = "u" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_coefficientcontainer(restriction)
    evaluate_coefficient(element, name, container, restriction, component)
    return name


def name_apply_function_gradient(element, restriction, component):
    rawname = "gradz_func" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient_gradient(element, name, container, restriction, component)
    return name


def name_apply_function(element, restriction, component):
    rawname = "z_func" + "_".join(str(c) for c in component)
    name = restricted_name(rawname, restriction)
    container = name_applycontainer(restriction)
    evaluate_coefficient(element, name, container, restriction, component)
    return name


def name_testfunctionspace(restriction):
    return restricted_name("lfsv", restriction)


def name_trialfunctionspace(restriction):
    return restricted_name("lfsu", restriction)


def type_testfunctionspace():
    return "LFSV"


def type_trialfunctionspace():
    return "LFSU"


def name_coefficientcontainer(restriction):
    name = restricted_name("x", restriction)
    return name


def name_applycontainer(restriction):
    name = restricted_name("z", restriction)
    return name


@pymbolic_expr
def pymbolic_coefficient(container, lfs, index):
    # TODO introduce a proper type for local function spaces!
    if isinstance(lfs, str):
        valuearg(lfs, dtype=loopy.types.NumpyType("str"))

    # If the LFS is not yet a pymbolic expression, make it one
    from pymbolic.primitives import Expression
    if not isinstance(lfs, Expression):
        lfs = Variable(lfs)

    from dune.perftool.loopy.functions import CoefficientAccess
    return Call(CoefficientAccess(container), (lfs, Variable(index),))


def type_coefficientcontainer():
    return "X"


def name_argumentspace(ma):
    from ufl.classes import Argument, Coefficient
    if isinstance(ma.argexpr, Argument):
        if ma.argexpr.number() == 0:
            return name_testfunctionspace(ma.restriction)
        if ma.argexpr.number() == 1:
            return name_trialfunctionspace(ma.restriction)
    if isinstance(ma.argexpr, Coefficient):
        # Index 0 is reserved for trialfunction, index 1 is reserved for jacobian apply function
        assert ma.argexpr.count() < 2
        return name_trialfunctionspace(ma.restriction)
    # We should never encounter an argument other than 0 or 1
    assert False


def name_jacobian(restriction1, restriction2):
    # Restrictions may only differ if NONE
    if (restriction1 == Restriction.NONE) or (restriction2 == Restriction.NONE):
        assert restriction1 == restriction2
    return restricted_name(restricted_name("jac", restriction1), restriction2)


def type_jacobian():
    return "J"


def name_residual(restriction):
    return restricted_name("r", restriction)


def type_residual():
    return "R"


def name_accumulation_variable(restrictions=(Restriction.NONE,)):
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        return name_residual(*restrictions)
    if ft == 'jacobian':
        return name_jacobian(*restrictions)
    assert False


def type_accumulation_variable():
    from dune.perftool.generation import get_global_context_value
    ft = get_global_context_value("form_type")
    if ft == 'residual' or ft == 'jacobian_apply':
        return type_residual()
    if ft == 'jacobian':
        return type_jacobian()
    assert False
